@@ -52,6 +52,7 @@ package com.openexchange.filestore.swift;
 import static com.openexchange.osgi.Tools.requireService;
 import java.net.URI;
 import java.net.URISyntaxException;
+import java.util.ArrayList;
 import java.util.LinkedList;
 import java.util.List;
 import java.util.concurrent.Callable;
@@ -64,19 +65,32 @@ import java.util.regex.Matcher;
 import java.util.regex.Pattern;
 import org.apache.http.client.HttpClient;
 import org.apache.http.client.utils.URIBuilder;
+import org.json.JSONArray;
+import org.json.JSONException;
+import org.json.JSONObject;
 import com.openexchange.config.ConfigurationService;
 import com.openexchange.configuration.ConfigurationExceptionCodes;
+import com.openexchange.database.DatabaseService;
 import com.openexchange.exception.OXException;
+import com.openexchange.filestore.DatabaseAccess;
+import com.openexchange.filestore.DatabaseAccessService;
 import com.openexchange.filestore.FileStorage;
+import com.openexchange.filestore.FileStorageInfo;
+import com.openexchange.filestore.FileStorageInfoService;
 import com.openexchange.filestore.FileStorageProvider;
+import com.openexchange.filestore.FileStorages;
 import com.openexchange.filestore.swift.chunkstorage.ChunkStorage;
 import com.openexchange.filestore.swift.chunkstorage.RdbChunkStorage;
 import com.openexchange.filestore.swift.impl.AuthInfo;
+import com.openexchange.filestore.swift.impl.ConfigType;
 import com.openexchange.filestore.swift.impl.EndpointPool;
 import com.openexchange.filestore.swift.impl.SwiftClient;
 import com.openexchange.filestore.swift.impl.SwiftConfig;
+import com.openexchange.filestore.swift.impl.TokenAndResponse;
+import com.openexchange.filestore.swift.impl.token.Token;
 import com.openexchange.filestore.swift.impl.token.TokenStorage;
 import com.openexchange.filestore.swift.impl.token.TokenStorageImpl;
+import com.openexchange.filestore.utils.DefaultDatabaseAccess;
 import com.openexchange.java.Strings;
 import com.openexchange.rest.client.httpclient.HttpClients;
 import com.openexchange.rest.client.httpclient.HttpClients.ClientConfig;
@@ -137,16 +151,31 @@ public class SwiftFileStorageFactory implements FileStorageProvider {
 
                     // Extract context and user from URI path
                     String filestoreID = extractFilestoreID(uri);
-                    int[] contextAndUser = extractContextAndUser(uri);
-                    int contextId = contextAndUser[0];
-                    int userId = contextAndUser[1];
-                    LOG.debug("Using \"{}\" as filestore ID, context ID of filestore is \"{}\", user ID is \"{}\".", filestoreID, contextId, userId);
+                    ExtractionResult extractionResult = extractFrom(uri);
+
+                    DatabaseAccess databaseAccess;
+                    int contextId;
+                    int userId;
+                    if (extractionResult.hasContextUserAssociation()) {
+                        contextId = extractionResult.getContextId();
+                        userId = extractionResult.getUserId();
+                        databaseAccess = new DefaultDatabaseAccess(userId, contextId, services.getService(DatabaseService.class));
+                        LOG.debug("Using \"{}\" as filestore ID, context ID of filestore is \"{}\", user ID is \"{}\".", filestoreID, contextId, userId);
+                    } else {
+                        contextId = 0;
+                        userId = 0;
+                        databaseAccess = lookUpDatabaseAccess(uri, extractionResult.getPrefix());
+                    }
+
+                    // Ensure required tables do exist
+                    databaseAccess.createIfAbsent(RdbChunkStorage.getRequiredTables());
 
                     // Initialize file storage using dedicated client & chunk storage
-                    SwiftClient client = initClient(filestoreID, contextId, userId);
-                    ChunkStorage chunkStorage = new RdbChunkStorage(services, contextId, userId);
+                    SwiftClient client = initClient(filestoreID, extractionResult.getPrefix());
+                    ChunkStorage chunkStorage = new RdbChunkStorage(databaseAccess, contextId, userId);
                     return new SwiftFileStorage(client, chunkStorage);
                 }
+
             });
 
             f = storages.putIfAbsent(uri, ft);
@@ -171,6 +200,19 @@ public class SwiftFileStorageFactory implements FileStorageProvider {
         }
     }
 
+    static DatabaseAccess lookUpDatabaseAccess(URI uri, String prefix) throws OXException {
+        FileStorageInfoService infoService = FileStorages.getFileStorageInfoService();
+        FileStorageInfo info = infoService.getFileStorageIdFor(uri);
+
+        DatabaseAccessService databaseAccessService = FileStorages.getDatabaseAccessService();
+        DatabaseAccess databaseAccess = databaseAccessService.getAccessFor(info.getId(), prefix);
+        if (null == databaseAccess) {
+            throw new IllegalArgumentException("No database access for file storage " + info.getId() + " with prefix \"" + prefix + "\"");
+        }
+
+        return databaseAccess;
+    }
+
     @Override
     public FileStorage getInternalFileStorage(URI uri) throws OXException {
         return getFileStorage(uri);
@@ -209,7 +251,7 @@ public class SwiftFileStorageFactory implements FileStorageProvider {
      * @return The extracted context and user identifiers
      * @throws IllegalArgumentException If URI's path does not follow expected pattern
      */
-    int[] extractContextAndUser(URI uri) {
+    ExtractionResult extractFrom(URI uri) {
         String path = uri.getPath();
         while (0 < path.length() && '/' == path.charAt(0)) {
             path = path.substring(1);
@@ -222,26 +264,54 @@ public class SwiftFileStorageFactory implements FileStorageProvider {
             if (false == matcher.matches()) {
                 throw new IllegalArgumentException("Path does not match the expected pattern \"\\d+_ctx_store\"");
             }
-            return new int[] {Integer.parseInt(matcher.group(1)), 0};
+            return new ExtractionResult(0, Integer.parseInt(matcher.group(1)));
         }
 
-        // Expect user store identifier
-        Matcher matcher = USER_STORE_PATTERN.matcher(path);
-        if (false == matcher.matches()) {
-            throw new IllegalArgumentException("Path does not match the expected pattern \"(\\d+)_ctx_(\\d+)_user_store\"");
+        if (path.endsWith("user_store")) {
+            // Expect user store identifier
+            Matcher matcher = USER_STORE_PATTERN.matcher(path);
+            if (false == matcher.matches()) {
+                throw new IllegalArgumentException("Path does not match the expected pattern \"(\\d+)_ctx_(\\d+)_user_store\"");
+            }
+            return new ExtractionResult(Integer.parseInt(matcher.group(2)), Integer.parseInt(matcher.group(1)));
         }
-        return new int[] {Integer.parseInt(matcher.group(1)), Integer.parseInt(matcher.group(2))};
+
+        // Any path that serves as prefix; e.g. "photos"
+        return new ExtractionResult(sanitizePathForPrefix(path, uri));
+    }
+
+    private static String sanitizePathForPrefix(String path, URI uri) {
+        if (Strings.isEmpty(path)) {
+            throw new IllegalArgumentException("Path is empty in URI: " + uri);
+        }
+
+        StringBuilder sb = null;
+        for (int k = path.length(), i = 0; k-- > 0; i++) {
+            char ch = path.charAt(i);
+            if ('_' == ch) {
+                // Underscore not allowed
+                if (null == sb) {
+                    sb = new StringBuilder(path.length());
+                    sb.append(path, 0, i);
+                }
+            } else {
+                // Append
+                if (null != sb) {
+                    sb.append(ch);
+                }
+            }
+        }
+        return null == sb ? path : sb.toString();
     }
 
     /**
      * Initializes an {@link SwiftClient} as configured by the referenced authority part of the supplied end-points.
      *
      * @param uri The file storage identifier
-     * @param contextID The context identifier
-     * @param userID The user identifier
+     * @param prefix The prefix to use
      * @return The client
      */
-    SwiftClient initClient(String filestoreID, int contextID, int userID) throws OXException {
+    SwiftClient initClient(String filestoreID, String prefix) throws OXException {
         SwiftConfig swiftConfig = swiftConfigs.get(filestoreID);
         if (swiftConfig == null) {
             SwiftConfig newSwiftConfig = initSwiftConfig(filestoreID);
@@ -253,7 +323,7 @@ public class SwiftFileStorageFactory implements FileStorageProvider {
             }
         }
 
-        return new SwiftClient(swiftConfig, contextID, userID, tokenStorage);
+        return new SwiftClient(swiftConfig, prefix, tokenStorage);
     }
 
     /**
@@ -276,16 +346,21 @@ public class SwiftFileStorageFactory implements FileStorageProvider {
         }
         String authValue = requireProperty(filestoreID, "authValue", config);
 
+        // Config type
+        ConfigType configType;
+        {
+            String sConfigType = config.getProperty(property(filestoreID, "configType"), ConfigType.MANUAL.getId());
+            configType = ConfigType.configTypeFor(sConfigType);
+            if (null == configType) {
+                throw ConfigurationExceptionCodes.INVALID_CONFIGURATION.create("Invalid value for 'com.openexchange.filestore.swift." + filestoreID + ".configType': " + sConfigType);
+            }
+        }
+
         // Tenant name, identity URL, and domain
         String tenantName = optProperty(filestoreID, "tenantName", config);
         String identityUrl = optProperty(filestoreID, "identityUrl", config);
         String domain = optProperty(filestoreID, "domain", config);
 
-        // End-point configuration
-        String protocol = requireProperty(filestoreID, "protocol", config);
-        String path = requireProperty(filestoreID, "path", config);
-        String hosts = requireProperty(filestoreID, "hosts", config);
-
         // HTTP client configuration
         int maxConnections = config.getIntProperty(property(filestoreID, "maxConnections"), 100);
         int maxConnectionsPerHost = config.getIntProperty(property(filestoreID, "maxConnectionsPerHost"), 100);
@@ -293,45 +368,102 @@ public class SwiftFileStorageFactory implements FileStorageProvider {
         int socketReadTimeout = config.getIntProperty(property(filestoreID, "socketReadTimeout"), 15000);
         int heartbeatInterval = config.getIntProperty(property(filestoreID, "heartbeatInterval"), 60000);
 
-        List<String> urls = new LinkedList<String>();
-        for (String host : Strings.splitAndTrim(hosts, ",")) {
-            URIBuilder uriBuilder = new URIBuilder().setScheme(protocol);
-            String[] hostAndPort = Strings.splitByColon(host);
-            if (hostAndPort.length == 1) {
-                uriBuilder.setHost(host);
-            } else if (hostAndPort.length == 2) {
-                try {
-                    uriBuilder.setHost(hostAndPort[0]).setPort(Integer.parseInt(hostAndPort[1]));
-                } catch (NumberFormatException e) {
-                    throw ConfigurationExceptionCodes.INVALID_CONFIGURATION.create(e, "Invalid value for 'com.openexchange.filestore.swift." + filestoreID + ".hosts': " + hosts);
-                }
-            } else {
-                throw ConfigurationExceptionCodes.INVALID_CONFIGURATION.create("Invalid value for 'com.openexchange.filestore.swift." + filestoreID + ".hosts': " + hosts);
-            }
-
-            uriBuilder.setPath(path);
-            try {
-                String baseUrl = uriBuilder.build().toString();
-                if (baseUrl.endsWith("/")) {
-                    baseUrl = baseUrl.substring(0, baseUrl.length() - 1);
-                }
-                urls.add(baseUrl);
-            } catch (URISyntaxException e) {
-                throw ConfigurationExceptionCodes.INVALID_CONFIGURATION.create(e, "Swift configuration leads to invalid URI: " + uriBuilder.toString());
-            }
-        }
-
-        if (urls.isEmpty()) {
-            throw ConfigurationExceptionCodes.INVALID_CONFIGURATION.create("Invalid value for 'com.openexchange.filestore.swift." + filestoreID + ".hosts': " + hosts);
-        }
-
+        // Create the HTTP client
         HttpClient httpClient = HttpClients.getHttpClient(ClientConfig.newInstance()
             .setMaxTotalConnections(maxConnections)
             .setMaxConnectionsPerRoute(maxConnectionsPerHost)
             .setConnectionTimeout(connectionTimeout)
             .setSocketReadTimeout(socketReadTimeout));
-        EndpointPool endpointPool = new EndpointPool(filestoreID, urls, httpClient, heartbeatInterval, requireService(TimerService.class, services));
-        return new SwiftConfig(filestoreID, userName, new AuthInfo(authValue, authType, tenantName, domain, identityUrl), httpClient, endpointPool);
+
+        // Create the auth info
+        AuthInfo authInfo = new AuthInfo(authValue, authType, tenantName, domain, identityUrl);
+
+        // End-points...
+        EndpointPool endpointPool;
+        if (ConfigType.MANUAL == configType) {
+            // Manual end-point configuration
+            String protocol = requireProperty(filestoreID, "protocol", config);
+            String path = requireProperty(filestoreID, "path", config);
+            String hosts = requireProperty(filestoreID, "hosts", config);
+
+            List<String> urls = new LinkedList<String>();
+            for (String host : Strings.splitAndTrim(hosts, ",")) {
+                URIBuilder uriBuilder = new URIBuilder().setScheme(protocol);
+                String[] hostAndPort = Strings.splitByColon(host);
+                if (hostAndPort.length == 1) {
+                    uriBuilder.setHost(host);
+                } else if (hostAndPort.length == 2) {
+                    try {
+                        uriBuilder.setHost(hostAndPort[0]).setPort(Integer.parseInt(hostAndPort[1]));
+                    } catch (NumberFormatException e) {
+                        throw ConfigurationExceptionCodes.INVALID_CONFIGURATION.create(e, "Invalid value for 'com.openexchange.filestore.swift." + filestoreID + ".hosts': " + hosts);
+                    }
+                } else {
+                    throw ConfigurationExceptionCodes.INVALID_CONFIGURATION.create("Invalid value for 'com.openexchange.filestore.swift." + filestoreID + ".hosts': " + hosts);
+                }
+
+                uriBuilder.setPath(path);
+                try {
+                    String baseUrl = uriBuilder.build().toString();
+                    if (baseUrl.endsWith("/")) {
+                        baseUrl = baseUrl.substring(0, baseUrl.length() - 1);
+                    }
+                    urls.add(baseUrl);
+                } catch (URISyntaxException e) {
+                    throw ConfigurationExceptionCodes.INVALID_CONFIGURATION.create(e, "Swift configuration leads to invalid URI: " + uriBuilder.toString());
+                }
+            }
+
+            if (urls.isEmpty()) {
+                throw ConfigurationExceptionCodes.INVALID_CONFIGURATION.create("Invalid value for 'com.openexchange.filestore.swift." + filestoreID + ".hosts': " + hosts);
+            }
+
+            endpointPool = new EndpointPool(filestoreID, urls, httpClient, heartbeatInterval, null, requireService(TimerService.class, services));
+        } else {
+            if (AuthInfo.Type.PASSWORD_V3 != authInfo.getType()) {
+                throw ConfigurationExceptionCodes.INVALID_CONFIGURATION.create("Invalid value for 'com.openexchange.filestore.swift." + filestoreID + ".configType': " + configType.getId() + ". Only supported for " + AuthInfo.Type.PASSWORD_V3.getId() + " auth type.");
+            }
+
+            // By interface, region and container name
+            String sInterface = requireProperty(filestoreID, "interface", config);
+            String region = requireProperty(filestoreID, "region", config);
+            String containerName = requireProperty(filestoreID, "containerName", config);
+
+            try {
+                TokenAndResponse tokenAndResponse = SwiftClient.doAcquireNewToken(userName, authInfo, httpClient);
+                Token initialToken = tokenAndResponse.getToken();
+                JSONArray jCatalog = tokenAndResponse.getJsonResponse().getJSONObject("token").getJSONArray("catalog");
+                JSONArray jEndpoints = null;
+                for (int k = jCatalog.length(), i = 0; null == jEndpoints && k-- > 0; i++) {
+                    JSONObject jCatalogEntry = jCatalog.getJSONObject(i);
+                    if ("swift".equals(jCatalogEntry.optString("name", null)) && "object-store".equals(jCatalogEntry.optString("type", null))) {
+                        jEndpoints = jCatalogEntry.getJSONArray("endpoints");
+                    }
+                }
+                if (null == jEndpoints) {
+                    throw ConfigurationExceptionCodes.INVALID_CONFIGURATION.create("No such catalog entry with \"name\"=\"swift\" and \"type\"=\"object-store\".");
+                }
+
+                int numOfEndpoints = jEndpoints.length();
+                List<String> urls = new ArrayList<String>(numOfEndpoints);
+                for (int k = numOfEndpoints, i = 0; k-- > 0; i++) {
+                    JSONObject jEndpoint = jEndpoints.getJSONObject(i);
+                    if (sInterface.equals(jEndpoint.optString("interface", null)) && region.equals(jEndpoint.optString("region", null))) {
+                        urls.add(jEndpoint.getString("url") + "/" + containerName);
+                    }
+                }
+
+                if (urls.isEmpty()) {
+                    throw ConfigurationExceptionCodes.INVALID_CONFIGURATION.create("No such end-points for \"interface\"=\"" + sInterface + "\" and \"region\"=\"" + region + "\".");
+                }
+
+                endpointPool = new EndpointPool(filestoreID, urls, httpClient, heartbeatInterval, initialToken, requireService(TimerService.class, services));
+            } catch (JSONException e) {
+                throw ConfigurationExceptionCodes.INVALID_CONFIGURATION.create(e, "Unexpected JSON response.");
+            }
+        }
+
+        return new SwiftConfig(filestoreID, userName, authInfo, httpClient, endpointPool);
     }
 
     private static String requireProperty(String filestoreID, String property, ConfigurationService config) throws OXException {
