@@ -106,53 +106,23 @@ public class SwiftClient {
     private final EndpointPool endpoints;
     private final HttpClient httpClient;
     private final String prefix;
-    private final int contextId;
-    private final int userId;
     private final TokenStorage tokenStorage;
 
     /**
      * Initializes a new {@link SwiftClient}.
      *
      * @param swiftConfig The Swift configuration
-     * @param contextId The context identifier
-     * @param userId The user identifier
+     * @param prefix The prefix to use
      * @param tokenStorage The token storage
      */
-    public SwiftClient(SwiftConfig swiftConfig, int contextId, int userId, TokenStorage tokenStorage) {
+    public SwiftClient(SwiftConfig swiftConfig, String prefix, TokenStorage tokenStorage) {
         super();
         this.userName = swiftConfig.getUserName();
         this.authValue = swiftConfig.getAuthInfo();
         this.endpoints = swiftConfig.getEndpointPool();
         this.httpClient = swiftConfig.getHttpClient();
-        this.contextId = contextId;
-        this.userId = userId;
         this.tokenStorage = tokenStorage;
-
-        // E.g. "57462ctx5userstore" or "57462ctxstore"
-        StringBuilder sb = new StringBuilder(32).append(contextId).append("ctx");
-        if (userId > 0) {
-            sb.append(userId).append("user");
-        }
-        sb.append("store");
-        prefix = sb.toString();
-    }
-
-    /**
-     * Gets the context identifier
-     *
-     * @return The context identifier
-     */
-    public int getContextId() {
-        return contextId;
-    }
-
-    /**
-     * Gets the user identifier
-     *
-     * @return The user identifier
-     */
-    public int getUserId() {
-        return userId;
+        this.prefix = prefix;
     }
 
     /**
@@ -600,7 +570,7 @@ public class SwiftClient {
      */
     private Endpoint getEndpoint() throws OXException {
         // Grab next end-point to use
-        Endpoint endpoint = endpoints.get(contextId, userId);
+        Endpoint endpoint = endpoints.get(prefix);
         if (endpoint == null) {
             throw SwiftExceptionCode.STORAGE_UNAVAILABLE.create();
         }
@@ -712,6 +682,19 @@ public class SwiftClient {
     }
 
     private Token doAcquireNewToken() throws OXException {
+        return doAcquireNewToken(userName, authValue, httpClient).getToken();
+    }
+
+    /**
+     * Acquires a new token according to specified arguments.
+     *
+     * @param userName The associated user name
+     * @param authValue The auth value to use
+     * @param httpClient The optional HTTP client
+     * @return The newly acquired token
+     * @throws OXException If token cannot be acquired
+     */
+    public static TokenAndResponse doAcquireNewToken(String userName, AuthInfo authValue, HttpClient httpClient) throws OXException {
         // Get a new one
         HttpPost post = null;
         HttpResponse response = null;
@@ -779,7 +762,7 @@ public class SwiftClient {
             int status = statusLine.getStatusCode();
             if (HttpServletResponse.SC_OK == status || HttpServletResponse.SC_CREATED == status) {
                 JSONObject jResponse = new JSONObject(new InputStreamReader(response.getEntity().getContent(), Charsets.UTF_8));
-                return authValue.getType().getParser().parseTokenFrom(jResponse, response);
+                return new TokenAndResponse(authValue.getType().getParser().parseTokenFrom(jResponse, response), jResponse);
             }
 
             String reasonPhrase = statusLine.getReasonPhrase();
